#!/usr/bin/perl -w
#######################################################################################
# monitorLoad.pl
# Written by: Brandon Zehm <caspian@dotconf.net>
# 
# Purpose:
#   To monitor the 15 minute system load average and print 
#   information to the system log.
#
#######################################################################################
use strict;

## Global Variable(s)
my %conf = (
    "programName"          => 'sys-mon-drraw.pl',
    "version"              => '1.0',
    "authorName"           => 'mrak',                    ## Information about the author or owner of this script.
    "authorEmail"          => 'mrak@redigy.cz',
    "hostname"             => $ENV{'HOSTNAME'},                  ## Used in printmsg() for all output.
    "syslogProgram"        => '/usr/bin/logger',
    
    "run"                  => 0,                                 ## Whether we should run script or not
    "debug"                => 0,
    "stdout"               => 0,
    "syslog"               => 1,                                 ## Syslog messages by default (--stdout disables this)
    "facility"             => "local3",
    "priority"             => "NOTICE",
    "file"                 => "",
    "name"                 => "",
    "dsname"                 => [],
    "values"                 => [],
    "load"                 => "",
    "message"                 => ""
);




#############################
##
##      MAIN PROGRAM
##
#############################

## Initialize
initialize();

## Process Command Line
processCommandLine();


## Get input 

if($conf{'file'}){
	for (my $counter = 0; $counter <= 10; $counter++, sleep(10)) {
	    quit("OS-ERROR => 10 consecutive errors while trying to read /proc/loadavg", 1) if ($counter >= 10);
	    open(FILE, "<$conf{'file'}") or next;
	    $conf{'input'} = <FILE>;
	    close FILE;
	    last if $conf{'input'};
	}
}else{
	$conf{'input'} = <STDIN>;	
}



## Extract the value we need 
@{$conf{'values'}}=split(/\s+/, $conf{'input'}, $conf{'dsname'});

$conf{'message'} = sprintf("DRRAW:%s ",$conf{'name'}); 

for(my $i=0; $i < @{$conf{'dsname'}}; $i++){ 
		$conf{'message'} .= sprintf(" %s:%.02f ", $conf{'dsname'}[$i], $conf{'values'}[$i]); 
}

$conf{'message'} .= sprintf(" Timestamp:%d ", time()); 
## Generate an output message
if ($conf{'syslog'}) {
            system( qq{$conf{'syslogProgram'} -p $conf{'facility'}.$conf{'priority'} -t $conf{'hostname'} "$conf{'programName'}: $conf{'message'}" });
}else{
	    printmsg($conf{'message'}, 0);
}

## Quit
quit("",0);






######################################################################
## Function:    help ()
##
## Description: For all those newbies ;) 
##              Prints a help message and exits the program.
## 
######################################################################
sub help {
print <<EOM;

$conf{'programName'}-$conf{'version'} by $conf{'authorName'} <$conf{'authorEmail'}>

Synopsis: 
  Checks the load average and logs it to the syslog daemon.

Usage:  $conf{'programName'} -rsv [-f] -N name -n name1:name2
  Read first -n space separated values from STDIN or file and send into the syslog.
  Required:
    -r                        	run script
    -N name			data group name  
    -n name1:name2:name3	data source names  
  Optional:
    -s                    	print only on stdout
    -f name                   	input file 
    -v                        verbosity - use multiple times for greater effect
  Example:
	cat /proc/loadavg | sys-mon-drraw.pl -s -r -N TEST -n test1:test2:test3

EOM
quit("", 1);
}




######################################################################
##  Function: initialize ()
##  
##  Does all the script startup jibberish.
##  
######################################################################
sub initialize {
    
    ## Set STDOUT to flush immediatly after each print  
    $| = 1;
    
    ## Intercept signals
    $SIG{'QUIT'}  = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
    $SIG{'INT'}   = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
    $SIG{'KILL'}  = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
    $SIG{'TERM'}  = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
    
    ## ALARM and HUP signals are not supported in Win32
    unless ($^O =~ /win/i) {
        $SIG{'HUP'}   = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
        $SIG{'ALRM'}  = sub { quit("$$ - EXITING: Received SIG$_[0]", 1); };
    }
    
    ## Make sure the path is sane
    $ENV{'PATH'} .= ":/bin:/usr/bin:/usr/local/bin:/sbin:/usr/sbin:/usr/local/sbin";
    
    ## Fixup $conf{'hostname'}
    if ($conf{'hostname'}) {
        $conf{'hostname'} = $conf{'hostname'};	# opet nechapu? (puvodni kod...)
        # $conf{'hostname'} =~ s/\..*$//;  	#chci cele hostname
    }
    else {
        $conf{'hostname'} = "unknown";
    }
    
    return(0);
}









######################################################################
##  Function: processCommandLine ()
##  
##  Processes command line storing important data in global var %conf
##  
######################################################################
sub processCommandLine {
  
    ############################
    ##  Process command line  ##
    ############################
  
    my @ARGS = @ARGV;
    my $numargv = @ARGS;
    my $counter = 0;
    for ($counter = 0; $counter < $numargv; $counter++) {
    
    
        if ($ARGS[$counter] =~ s/^-v+//i) {               ## Verbosity ##
            $conf{'debug'} += (length($&) - 1);
        }

        elsif ($ARGS[$counter] =~ /^-f$/) {                  ## Run
            $conf{'file'} = $ARGS[++$counter];
	    next;
        }
        elsif ($ARGS[$counter] =~ /^-N$/) {                  ## name
            $conf{'name'} = $ARGS[++$counter];
	    next;
        }
        elsif ($ARGS[$counter] =~ /^-n$/) {                  ## dsname
            @{$conf{'dsname'}} = split(/:/, $ARGS[++$counter]);
	    next;
        }
        
        elsif ($ARGS[$counter] =~ /^-r$/) {                  ## Run
            $conf{'run'} = 1;
        }
        elsif ($ARGS[$counter] =~ /^-s$/) {                  ## stdout 
            $conf{'stdout'} = 1;
            $conf{'syslog'} = 0;
        }
        
        elsif ($ARGS[$counter] =~ /^-h$|^--help$/i) {        ## Help ##
            help();
        }
        
        else {                                               ## The Message ##
            quit("OS-ERROR => The option [$ARGS[$counter]] is unknown. Try --help", 1);
        }

    }
    
    ## If the user didn't use a -r print the help
    if ($conf{'run'} == 0 || !$conf{'name'} || !@{$conf{'dsname'}} ) {
        help();
    }
    
  
    return(1);
}






###############################################################################################
##  Function:    printmsg (string $message, int $level)
##
##  Description: Handles all messages - 
##               Depending on the state of the program it will log
##               messages to a log file, print them to STDOUT or both.
##               
##
##  Input:       $message          A message to be printed, logged, etc.
##               $level            The debug level of the message. If not defined 0
##                                 will be assumed.  0 is considered a normal message, 
##                                 1 and higher is considered a debug message.
##  
##  Output:      Prints to STDOUT, to LOGFILE, both, or none depending 
##               on the state of the program and the debug level specified.
##  
##  Example:     printmsg("ERROR => The file could not be opened!", 0);
###############################################################################################
sub printmsg {
    ## Assign incoming parameters to variables
    my ( $message, $level ) = @_;
    
    ## Make sure input is sane
    $level = 0 if (!defined($level));
    
    ## Continue only if the debug level of the program is >= message debug level.
    if ($conf{'debug'} >= $level) {
        
        ## Get the date in the format: Dec  3 11:14:04
        my ($sec, $min, $hour, $mday, $mon) = localtime();
        $mon = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')[$mon];
        my $date = sprintf("%s %02d %02d:%02d:%02d", $mon, $mday, $hour, $min, $sec);
    
        ## Syslog the message is needed
        if ($conf{'syslog'}) {
            system( qq{$conf{'syslogProgram'} -p $conf{'facility'}.$conf{'priority'} -t $conf{'hostname'} "$conf{'programName'}: $message" });
        }
        
        ## Print to STDOUT always if debugging is enabled, or if conf{stdout} is true.
        if ( ($conf{'debug'} >= 1) or ($conf{'stdout'} == 1) ) {
            print "$date $conf{'hostname'} $conf{'programName'}: $message\n";
        }
    }
    
    ## Return 0 errors
    return(0);
}











######################################################################
##  Function:    quit (string $message, int $errorLevel)
##  
##  Description: Exits the program, optionally printing $message.  It 
##               returns an exit error level of $errorLevel to the 
##               system  (0 means no errors, and is assumed if empty.)
##
##  Example:     quit("Exiting program normally", 0);
######################################################################
sub quit {
  my %incoming = ();
  (
    $incoming{'message'},
    $incoming{'errorLevel'}
  ) = @_;
  $incoming{'errorLevel'} = 0 if (!defined($incoming{'errorLevel'}));
  
  
  ## Print exit message
  if ($incoming{'message'}) { 
    printmsg($incoming{'message'}, 0);
  }
  
  ## Exit
  exit($incoming{'errorLevel'});
}






