#!/bin/bash

PATH=/sbin:/bin:/usr/sbin:/usr/bin:/root/bin:/usr/local/sbin:/usr/local/bin

name=`uname -n`
fs_usage_warn=92
load_warn=3.5
min_uptime=400

################## tests ##################

# --- mark ---
t_mark() {
echo
	 logger -p local3.warn -t $name -- "--mark-- `date +%s`"
}

t_check(){
echo
	 logger -p local3.warn -t $name -- "special logdog check"
}



# load
t_load() {
 rload="$(uptime |awk -F'average:' '{ print $2}')"
  x="$(echo $rload | sed s/,//g | awk '{ print $2}')"
  y="$(echo "$x >= $load_warn" | bc)"
  if [ "$y" == "1" ];then   
  	rtotalprocess="$(ps axue | grep -vE "^USER|grep|ps" | wc -l)"
	#echo "$name Average load $rload running $rtotalprocess processes"
	logger -p local3.warn -t $name "Average load $rload running $rtotalprocess processes"
  fi
}

# disk free  	
t_disk-free() {
 df -h -t ext2 -t ext3 | grep -v ^Filesystem | while read ldf
 do
   p=$(echo $ldf|awk '{print $5 }')
   if [ ${p%'%'} -gt $fs_usage_warn ] ;then
	#echo "$name Disk usage $ldf"
	logger -p local3.warn -t $name  "$name Disk usage $ldf"
   fi
 done
}

# RAID monitor
t_disk-raid() {
 ps ax|grep -v 'grep' |grep 'mdadm --monitor' &>/dev/null
 ret1=$?
 grep '^PROGRAM.*mdadm-logger.sh' /etc/mdadm.conf &>/dev/null 
 ret2=$?
 if [ $ret1 -ne 0 -o $ret2 -ne 0 ];then
   mdadm -F -1 --scan -p /usr/local/sbin/system-mdadm-logger.sh
 fi

}


# chech uptime
t_uptime() {
  sec=$(cat /proc/uptime|awk '{print $1}')
  sec=${sec%%.*}
  if [ $sec -lt $min_uptime ];then
	#echo "Uptime too low: $sec s"
	logger -p local3.notice -t $name "Uptime is low ${sec}s"
  fi

}


###############      main      #######################
USAGE="Usage: $0 mark uptime load disk-free disk-raid"

while (($#));do

  case "$1" in
    mark)
        t_mark
        ;;
    check)
        t_check
        ;;
    uptime)
        t_uptime
        ;;
    load)
        t_load
        ;;
    disk-free)
        t_disk-free
        ;;
    disk-raid)
        t_disk-raid
        ;;
    *)
        echo "$USAGE"
        exit 1
	;;
  esac
shift
done
